// Logical functions for Verilog.

function  [31:0] identity;
      input  [31:0] n;
      begin
         identity = n;
      end
endfunction // identity

// Useful because brackets may not be used to subscript constants in Verilog.
function logbit_P;
      input [4:0] ind;
      input [31:0] int;
      begin
         logbit_P = ((1 << ind) & int) ? 1'b1 : 1'b0;
      end
endfunction // logbitp


// This is useful only if ISTART and IEND are not constants.
function [31:0] bit_field;
      input [31:0] n;
      input [4:0] istart;
      input [4:0] iend;
      begin
         bit_field = (n >> istart) & ((1 << (iend - istart)) - 1);
      end
endfunction // bit_field

function [31:0] copy_bit;
      input [4:0] ind;
      input [31:0] to;
      input [31:0] bool;
      reg  [31:0] mask;
      begin
         mask = 1 << ind;
         copy_bit = bool ? (to | mask) : (to & (~mask));
      end
endfunction // copy_bit

function  [31:0] bitwise_if;
      input  [31:0] mask;
      input  [31:0] n0;
      input  [31:0] n1;
      begin
         bitwise_if = (mask & n0) | ((~mask) & n1);
      end
endfunction // bitwise_if

function  [31:0] copy_bit_field;
      input  [31:0] to;
      input [4:0] istart;
      input [4:0] iend;
      input  [31:0] from;
      reg  [31:0] mask;
      begin
         mask = ((1 << (iend - istart)) - 1) << istart;
         copy_bit_field = bitwise_if(mask, from << istart, to);
      end
endfunction // copy_bit_field

/* -----\/----- EXCLUDED -----\/-----

function random_Percent;
      input percentage;
      reg rn;
      begin
         rn = 7'b1111111 & $random;
         while (rn > 99) rn = 7'b1111111 & $random;
         random_Percent = (rn < percentage);
      end
endfunction // random_Percent
 -----/\----- EXCLUDED -----/\----- */
