-- "readtext.vhd"; Extends "std.textio" function READ to std_ulogic_vector.

LIBRARY ieee;

USE ieee.std_logic_1164.ALL;
USE std.textio.all;

PACKAGE readtext IS

    procedure Read(L:inout LINE; VALUE: out std_ulogic; GOOD : out BOOLEAN);
    procedure Read(L:inout LINE; VALUE: out std_ulogic);

    procedure Read(L:inout LINE; VALUE: out std_ulogic_vector; GOOD : out BOOLEAN);
    procedure Read(L:inout LINE; VALUE: out std_ulogic_vector);

END readtext;


PACKAGE BODY readtext IS

    procedure Skip_white(variable L : in LINE; pos : inout integer)
    is
    begin
	while pos <= L'high loop
	    case L(pos) is
		when ' ' | HT  =>
		    pos := pos + 1;
		when others =>
		    exit;
	    end case;
	end loop;
    end;

    procedure Shrink_line(L : inout LINE; pos : in integer)
    is
	variable old_L : LINE := L;
    begin
	if pos > 1 then
	    L := new string'(old_L(pos to old_L'high));
	    Deallocate(old_L);
	end if;
    end;

    procedure Report_results(good : boolean; read_type : string)
    is
    begin
	assert good
	    report "Could not read type " & read_type & " from line."
	    severity error;
    end;

    -----------------------------------------------------------------
    -- Std_Ulogic reading
    -----------------------------------------------------------------
    procedure Read(L:inout LINE; VALUE: out std_ulogic; GOOD : out BOOLEAN)
    is
      variable pos : integer;
      variable ok  : boolean := TRUE;
    begin
      if L /= NULL then
        pos := L'left;
        Skip_white(L, pos);
        if pos <= L'right then
          CASE L(pos) IS
            WHEN 'U'|'u' => value := 'U';
            WHEN '0'     => value := '0';
            WHEN '1'     => value := '1';
            WHEN 'X'|'x' => value := 'X';
            WHEN 'L'|'l' => value := 'L';
            WHEN 'H'|'h' => value := 'H';
            WHEN 'W'|'w' => value := 'W';
            WHEN 'Z'|'z' => value := 'Z';
            WHEN '-'     => value := '-';
            WHEN OTHERS => value := 'X'; ok := FALSE;
          END CASE;
        end if;
      end if;

      GOOD := ok;
      if ok then
        Shrink_line(L, pos + 1);
      end if;
    end;

    procedure Read(L:inout LINE; VALUE: out std_ulogic)
    is
      variable GOOD : BOOLEAN;
    begin
      Read(L, VALUE, GOOD);
      Report_results(GOOD, "STD_ULOGIC");
    end;

    -----------------------------------------------------------------
    -- std_ulogic_vector reading
    -----------------------------------------------------------------

    procedure Read(L:inout LINE; VALUE: out std_ulogic_vector; GOOD : out BOOLEAN)
    is
      alias    val  : std_ulogic_vector(1 to VALUE'length) is VALUE;
      variable vpos : integer := 0;	-- Index of last valid bit in val.
      variable lpos : integer;	-- Index of next unused char in L.
    begin
      if L /= NULL then
        lpos := L'left;
        Skip_white(L, lpos);
        while lpos <= L'right and vpos < VALUE'length loop
          vpos := vpos + 1;
          CASE L(lpos) IS
            WHEN 'U'|'u' => val(vpos) := 'U';
            WHEN '0'     => val(vpos) := '0';
            WHEN '1'     => val(vpos) := '1';
            WHEN 'X'|'x' => val(vpos) := 'X';
            WHEN 'L'|'l' => val(vpos) := 'L';
            WHEN 'H'|'h' => val(vpos) := 'H';
            WHEN 'W'|'w' => val(vpos) := 'W';
            WHEN 'Z'|'z' => val(vpos) := 'Z';
            WHEN '-'     => val(vpos) := '-';
            WHEN OTHERS => EXIT;
          END CASE;
          lpos := lpos + 1;
        END LOOP;
      END IF;

      if vpos = VALUE'length then
        GOOD := TRUE;
        Shrink_line(L, lpos);
      else
        GOOD := FALSE;
      end if;
    end;

    procedure Read(L:inout LINE; VALUE: out std_ulogic_vector)
    is
	variable GOOD : BOOLEAN;
    begin
	Read(L, VALUE, GOOD);
	Report_results(GOOD, "STD_ULOGIC_VECTOR");
    end;

END readtext;
