;;;; "run.scm" coordinate compilation and simulation.   -*-scheme-*-
;;; Copyright (C) 1995, 1996, 1997, 2001 Aubrey Jaffer

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

(require 'common-list-functions)
(defvar *design* #f)

(define (design-file des str)
  (if (symbol? str) (set! str (symbol->string str)))
  (string-append (symbol->string des) str (scheme-file-suffix)))

(define simsynch-vicinity
  (let ((vic (program-vicinity)))
    (lambda () vic)))

(define *spew-output* (current-output-port))
(define (translate des . opts)
  (if (string? des) (set! des (string->symbol des)))
  (set! *design* des)
  (set! opts (if (null? opts) 'verilog (car opts)))
  (if (string? opts) (set! opts (string->symbol opts)))
  (require (design-file *design* ""))
  (solidify-database *board*)
  (require opts)
;;;      (trace-all "scm2mach.scm") (untrace out out-pins c-ify-symbol lookup compile-exp infix-compile-exp registered-input? numbers-in-expression?)
;;;      (trace-all "simsynch.scm")
  (fluid-let ((*spew-output*
	       (open-file (string-append (symbol->string des) "-spew.scm")
			  "w")))
    (translate-design *design*)
    (close-output-port *spew-output*))
  ;;(print-nets)
  )

(define (print-nets)
  (display
   "\nHere are the ASIC packages and the connections required between them:\n")
  (display "$PACKAGES\n")

  (for-each-ptag (lambda (name pld-type signature optcom)
		   (printf "%s! %s; ??\n" pld-type name)))
  (display "$NETS\n")
  ((pin-table 'for-each-row-in-order)
   (lambda (pin-row)
     (apply (lambda (ptag name pin-type pin-number output-enable paired-input defining-block)
	      (cond
	       ((eq? pin-type 'input)
		(let ((found? #f))
		  ((pin-table 'for-each-row-in-order)
		   (lambda (pin-row)
		     (cond (found?)
			   (else (set! found? #t)
				 (printf "%s;  %s.%a" name ptag pin-number)))
		     (printf " %s.%a" (car pin-row) (cadddr pin-row)))
		   ;; Find only pins in other chips
		   (lambda (ptg) (and (not (eq? ptg 'HID))
				      (not (eq? ptg ptag))))
		   ;;  with the same signal name
		   name)
		  (if found? (newline))))))
	    pin-row))
   ;; For-each actual pins
   (lambda (ptg) (not (eq? ptg 'HID))))
  (display "$END\n"))

(define (comp filename)
  (let ((outfile (string-append "exp-" filename)))
    (require 'pprint-file)
    (require 'defmacroexpand)
    (require 'schlep)
    (printf "expanding %s -> %s\n" filename outfile)
    (pprint-filter-file filename defmacro:expand* outfile)
    (let ((stdc __STDC__))
      (set! __STDC__ #t)
      (compile outfile)
      (set! __STDC__ stdc))))

(define (c-ify-sym sym)
  (string->symbol
   (list->string
    (map (lambda (c) (if (eqv? #\- c) #\_ c))
	 (string->list (symbol->string sym))))))

(define (sym-flatten sym)
  (let ((ans (list->string
	      (set-difference (string->list (symbol->string sym))
			      '(#\[ #\])))))
    (string->symbol (if (and (> (string-length ans) 4)
			     (or (string-ci=? "pre:" (substring ans 0 4))
				 (string-ci=? "pci-" (substring ans 0 4))))
			(substring ans 4 (string-length ans))
			ans))))

(define (path->ptag name)
  (let* ((len (string-length name)))
    (do ((i (+ -1 len) (+ -1 i)))
	((or (negative? i) (memv (string-ref name i) '(#\\ #\/ #\:)))
	 (string->symbol
	  (if (negative? i)
	      name
	      (substring name (+ 1 i) len)))))))

(define (compare-plc ptag plc munge-sym)
  (define UPPTAG (string-upcase (symbol->string ptag)))
  (define (pin-compare pin-row)
    (apply
     (lambda (ptg name pin-type pin-number output-enable paired-input defining-block)
       (cond
	((vectorized? (or paired-input name))
	 (let* ((bounds (vector-bounds name))
		(first (cadr bounds))
		(last (caddr bounds))
		(range (- last first))
		(incer (sign range))
		(pinlist (if (and (list? pin-number)
				  (positive? incer))
			     (reverse pin-number)
			     pin-number)))
	   (do ((i 0 (+ 1 i))
		(j first (+ incer j)))
	       ((> i (abs range)))
	     (pin-compare
	      (list ptg
		    (refvectorize name j)
		    pin-type
		    (and (list? pinlist) (list-ref pinlist i)) ;pin-number
		    output-enable
		    (and paired-input (refvectorize paired-input j)) ;paired-input
		    )))))
	(else
	 (let* ((piname (munge-sym (or paired-input name)))
		(plc-row (assq piname plc)))
	   (cond ((and (not plc-row) (not paired-input)))
		 ((not plc-row)
		  (print 'no-pin-defined pin-row '<> plc-row))
		 ((eqv? pin-number (cadr plc-row)))
		 (else (printf "%S:%-15s changed from %3d to %3d\n"
			       UPPTAG name
			       pin-number (cadr plc-row))))))))
     pin-row))
  ;;(trace vector-bounds pin-compare refvectorize)
  ((pin-table 'for-each-row-in-order) pin-compare ptag))

#+foo           (sort
		      (lambda (rowa rowb)
			(string<? (symbol->string (car rowa))
				  (symbol->string (car rowb)))))

(define (compare-plcs des)
  (set! *design* des)
  (require (design-file *design* ""))
  (require 'line-i/o)
  (provide 'comparing-plc)
  (for-each-ptag
   (lambda (name pld-type signature optcom)
     (let* ((path (string-append
		   (symbol->string des) "-" (symbol->string name) ".plc"))
	    (ptag (path->ptag (symbol->string name)))
	    (plc '()))
       (define (update-plc row) (set! plc (cons row plc)))
       (cond ((file-exists? path)
	      (call-with-input-file path
		(lambda (ip)
		  ;; Read through file header
		  (cond ((not (eqv? 'machxl (read ip)))
			 (slib:error "not a MACHXL file" fname))
			(else (read-line ip)
			      (read-line ip)))
		  (do ((pinode (read ip) (read ip)))
		      ((eqv? pinode 'node))
		    (let* ((pinum (read ip))
			   (piname (read ip)))
		      (read-line ip)
		      (update-plc (list piname pinum))))))
	      (compare-plc ptag plc c-ify-sym))
	     (else (slib:error 'compare-plcs 'file-not-found path)))))))

;;; M-x compile:
;; scm -rrun -e"(compare-pin 'cz 'xp \"../cz/xp466\")"
;; scm -rrun -e"(compare-pin 'cdl 'xp \"../cdl/xpasic\")"

(define (signal-default sym)
  (define s (symbol->string sym))
  (cond ((vectorized? sym) "0")
	((and (> (string-length s) 1)
	      (memv (string-ref s (+ -1 (string-length s))) '(#\# #\- #\_)))
	 "#t")
	(else "#f")))

(define (compare-pin des ptag path)
  (set! *design* des)
  (require (design-file *design* ""))
  ;;(provide 'comparing-pin)
  (let ((pin '()))
    (define (update-pin row) (set! pin (cons row pin)))
    (call-with-input-file (string-append path ".pin")
      (lambda (iport)
	(define shape (read iport))
	(cond
	 ((not (list? shape)) (slib:error "not valid format" shape))
	 ((memq (car shape) '(pqfp plcc lcc soj))
	  (set-car! (cdr shape) (expand-range (cadr shape)))
	  (do ((rows (cadr shape) (cdr rows)))
	      ((null? rows))
	    (let* ((pnum (read iport))
		   (pnam (read iport))
		   (ptyp (read iport))
		   (p-type (cdr (or (assq ptyp
					  '((out . output)
					    (i/o . input/output)
					    (hiZ . input/output)
					    ;;(wire-and)
					    ;;(wire-or)
					    (clk . clock)
					    (in . input)
					    (v . unused)
					    ;;(internal)
					    ))
				    (cons ptyp ptyp)))))
	      (cond ((not (equal? pnum (car rows)))
		     (slib:error "pins out of sequence"
				 (car rows) "~=" pnum)))
	      (update-pin (list pnam pnum p-type))
	      (cond
	       ((memq pnam '(_i _c _ - gnd vcc tck tdo tms tdi trst-)))
	       ((pindex:row-retrieve ptag pnum)
		=> (lambda (prow)
		     (apply
		      (lambda (ptag pin-number symbol-name
				    pin-type signal-name)
			(cond
			 ((not (eq? pnam (sym-flatten signal-name)))
			  (printf
			   "pin %d signal names don't match: %s <> %s\n"
			   pnum pnam (sym-flatten signal-name)))
			 ((and (not (eq? p-type pin-type))
			       (not (and (eq? p-type 'input/output)
					 (or (eq? pin-type 'output)
					     (eq? pin-type 'input)))))
			  (printf
			   "pin %d (%s) signal types don't match: %s <> %s\n"
			   pnum pnam p-type pin-type))))
		      prow)))
	       (else
		(case p-type
		  ((output)
		   (printf "(synch:define %d %s:%s %s #f #t)\n" pnum ptag pnam
			   (signal-default pnam))
		   (printf "(synch:set! %s %s)\n" pnam pnam))
		  ((input/output)
		   (printf "(synch:define %d %s:%s %s #f #f)\n" pnum ptag pnam
			   (signal-default pnam))
		   (printf "(synch:set! %s %s)\n" pnam pnam))
		  ((input)
		   (printf "(synch:definput %d %s:%s %s)\n" pnum ptag pnam
			   (signal-default pnam)))
		  (else (printf "no define for pin %d %s\n" pnum pnam)))
		)))))
	 (else (slib:error "unknown format" shape)))))
    ;;(pretty-print pin)
    (compare-plc ptag pin sym-flatten)))

;;  ================================================================
;;;;		 Output state history in binary format.

(define dump:port #f)
(define *dump-list* #f)
(define *dump-widths* #f)

(defmacro dump:verilog args
  (set! *dump-list* args)
  `(spew:verilog ,@args))

(define dump:address 0)
(define (spew:verilog dump-file . args)
  (define len (length args))
  (set! *dump-list* args)
  (cond ((not dump:port)
	 (printf " Opening simulation dump file \"%s\"\n" dump-file)
	 (set! dump:port (open-file dump-file "wb"))))
  (display #\@ dump:port)
  (display (number->string dump:address 16) dump:port)
  (set! dump:address (+ dump:address len))
  (map (lambda (x)
	 (display #\space dump:port)
	 (display (case x
		    ((#t) #\1)
		    ((#f) #\0)
		    (else (cond ((not (number? x))
				 (report 'dump:verilog funny-value x) #\z)
				((eqv? *xxxx* x) #\x)
				((negative? x) #\z)
				(else (number->string x 16)))))
		  dump:port))
       args)
  (newline dump:port))

(defmacro dump:vhdl args
  (for-each (lambda (sig) (mark-uses 'spew sig)) (cdddr args))
  (set! *dump-list* args)
  (set! *dump-widths*
	(map (lambda (sym)
	       (cond ((vector-bounds sym)
		      => (lambda (bounds)
			   (abs (- (caddr bounds) (cadr bounds)))))
		     (else 0)))
	     (cdr (cddddr args))))
  `(spew:vhdl synch:count ,@args))

(define (spew:prin-header frequency names widths)
  (set! widths (map (lambda (x) (+ 1 x)) widths))
  (let* ((ss (map (lambda (i)
		    (string-upcase!
		     (cond ((string? i) i)
			   ((char? i) "")
			   (else (object->string i)))))
		  names))
	 (sl (map string-length ss))
	 (ml (apply max sl))
	 (ll (map (lambda (s l)
		    (string->list
		     (if (zero? l) (make-string ml #\nul)
			 (string-append s (make-string (- ml l) #\space)))))
		  ss sl))
	 (rl (apply map list ll))
	 (hht (length rl)))
    (for-each (lambda (str)
		(define linum (+ -1 (port-line dump:port)))
		(fprintf dump:port (string-append str " %5d") (- linum hht 1))
		(if frequency (fprintf dump:port " @ %Ks"
				       (/ (+ 1/2 linum) frequency)))
		(newline dump:port))
	      (map (lambda (l)
		     (apply
		      string-append
		      "." (if (char=? #\nul (car l)) "" (string (car l)))
		      (map (lambda (c width)
			     (define half (quotient (+ 1 width) 2))
			     (string-append
			      (make-string half #\space)
			      (string c)
			      (make-string (- width half) #\space)))
			   (cdr l)
			   (cdr widths))))
		   rl))
    (fprintf dump:port ".\\n")))

(define (spew:vhdl cycle:count dump-file . args)
  (define frequency (cadr args))
  (cond ((not dump:port)
	 (printf " Opening simulation dump file \"%s\"\n" dump-file)
	 (set! dump:port (open-file dump-file (_tracked "wb")))
	 (spew:prin-header
	  frequency
	  (map (lambda (sym)
		 (cond ((vector-bounds sym) => car) (else sym)))
	       (cdr (cddddr *dump-list*)))
	  *dump-widths*)))
  (map (lambda (x wid)
	 (display #\space dump:port)
	 (display
	  (cond ((eqv? 0 wid)
		 (case x
		   ((#t) #\1)
		   ((#f) #\0)
		   (else (cond ((not (number? x))
				(report 'dump:vhdl 'funny-value x) #\z)
			       ((eqv? *xxxx* x) #\U)
			       ((negative? x) #\Z)
			       (else (number->string x 2))))))
		((not (number? x))
		 (report 'dump:vhdl 'funny-value x)
		 (make-string (+ 1 wid) #\x))
		((eqv? *xxxx* x) (make-string (+ 1 wid) #\u))
		((negative? x) (make-string (+ 1 wid) #\z))
		(else (sprintf #f "%0*b" (+ 1 wid) x)))
	  dump:port))
       (cddddr args)
       *dump-widths*)
  (let ((linum (+ -1 (port-line dump:port))))
    (fprintf dump:port " %5d" cycle:count)
    (if frequency (fprintf dump:port " @ %Ks" (/ (+ 1/2 linum) frequency))))
  (newline dump:port))

(define (spew-list block frequency clock reset inputs inouts outputs)
  (define merged
    (map (lambda (sig)
	   (case (length sig)
	     ((1) (car sig))
	     ((4)
	      (set! sig (cadddr sig))
	      (case (length sig)
		((1) (car sig))
		(else `(logior
			,@(map (lambda (sig)
				 `(ash ,sig ,(caddr (vector-bounds sig))))
			       (reverse (map string->symbol
					     (sort (map symbol->string sig)
						   chap:string<?))))))))
	     (else (report 'funny-sig sig))))
	 (remove-duplicates (append inputs inouts outputs))))
  (list* block frequency clock reset (remove clock (remove reset merged))))

(define (generate-spew-lists)
  ((block-table 'for-each-row-in-order)
   (lambda (block-row)
     (apply
      (lambda (block-name clock reset clock-frequency synchronous-reset? optcom)
	;;(print 'module block-name 'clock clock 'reset reset)
	;;(set! inputs (merge-bounds (list clock) inputs))
	;;(set! inputs (merge-bounds (list reset) inputs))
	(fprintf *spew-output* ";; State dump for %s %s\n" *design* block-name)
	(call-with-classified-pins
	 #f block-name
	 (lambda (outputs inouts inputs registers)
	   (pretty-print
	    `(dump:vhdl
	      ,(string-append (symbol->string *design*) "-spew.dat")
	      ,@(spew-list
		 block-name clock-frequency clock reset inputs inouts outputs))
	    *spew-output*)))
	(newline *spew-output*))
      block-row))))
;;  ================================================================

(define (kb:announce-polling action)
  (cond ((and *interactive* (char-ready?))
	 (read-char)
	 (display (if (provided? 'char-ready?)
		      "Type any key to "
		      "Type CTRL-C to "))
	 (display action)
	 (newline))))
(define char-ready?
  (if (provided? 'char-ready?)
      char-ready?
      (lambda args #f)))
(define (kb:hit?)
  (and *interactive* (char-ready?)
       (cond ((eof-object? (peek-char))
	      (read-char)
	      (or (not (char-ready?))
		  (not (eof-object? (peek-char)))))
	     (else #t))))

(define (kb:break . args)
  (stack-trace)
  (require 'debug)
  (if (not (null? args)) (apply slib:warn args))
  (printf " * Breakpoint established: (bk) to resume.\n")
  (let ((val (breakpoint)))
    (cond ((char-ready?) (read-char)))
    val))
(define (bk)
  (require 'break)
  (synch:reset)
  (display "resuming...") (newline)
  (cond ((char-ready?) (read-char)))
  (continue))

(define (kb:break-on-hit . args)
  (and *interactive*
       ;;(not *load-pathname*)
       (char-ready?)
       (let ((pchr (peek-char)))
	 (and (not (eof-object? pchr))
	      (if (null? args)
		  (kb:break "" (case pchr
				 ((#\newline) "<newline>")
				 ((#\space) "<space>")
				 (else (string pchr)))
			    'key 'hit.)
		  (apply kb:break args))))))

;; DESIGN is a symbol.
;; NETLIST is a filename, optionally with ".net" suffix assumed.
;; PTAG-MAP is a filename.
(define (print-netlist-connections design netlist ptag-map)
  (set! *design* design)
  (require design)
  (db:read-u-ptag ptag-map)
  (db:read-net netlist)
  (letrec ((pin-retrieve (pin-table 'row:retrieve))
	   (print1 (lambda (ptag signame)
		     (let ((pin-row (pin-retrieve ptag signame)))
		       (display "          ")
		       (display signame)
		       (display " ")
		       (display (and pin-row
				     (list-ref pin-row 4))) ;output-enable
		       (display " ")
		       (newline)))))
    ((netlist-aggregate 'for-each-row-in-order)
     (lambda (row)
       (let ((connects (apply db:netlist-connects row)))
	 (cond ((pair? connects)
		(display row)
		(display ":\n")
		(for-each (lambda (row)
			    (apply print1 row)) connects))))))
    ((netlist-non-aggregate 'for-each-row-in-order)
     (lambda (row)
       (let ((connects (db:netlist-connects (car row) #f)))
	 (cond ((pair? connects)
		(display (car row))
		(display ":\n")
		(for-each (lambda (row)
			    (apply print1 row)) connects))))))))

