;;;; "simsynch.scm" simulate synchronous systems.	-*-scheme-*-
;;; Copyright (C) 1995, 1996, 1997, 2001, 2002 Aubrey Jaffer

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation, either version 3 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

(display ";			 loading SIMSYNCH ")
(display
 (call-with-input-file (in-vicinity (program-vicinity) "version.txi")
   (lambda (port) (read port) (read port) (read port))))
(newline)
(require 'database-browse)
(define b browse)

;;(provide 'check-turnaround)

(require 'sort)
(require 'array)
(require 'line-i/o)
(require 'string-case)
(require 'database-commands)
(require 'databases)
(require 'common-list-functions)
(require 'logic)			;(in-vicinity (program-vicinity) "logic.scm")
(require 'models)			;(in-vicinity (program-vicinity) "models.scm")

(define (create-board board-name dbtype)
  (define fname (string-append (symbol->string board-name) ".db"))
  (let ((boardb (create-database fname dbtype)))
    (cond ((not boardb)
	   (cond ((file-lock-owner fname)
		  (slib:error fname 'locked 'by (file-lock-owner fname))))
	   (slib:error 'could 'not 'create-database
		       (string-append (symbol->string board-name) ".db")
		       dbtype)))
    (set! boardb (add-command-tables boardb))
    (define-domains boardb
      '(optstring #f (lambda (x) (or (not x) (string? x))) string #f))
    (define-tables boardb
      '(constants
	((name symbol))
	((value expression)
	 (optcom optstring))
	())

      '(ptag
	((name symbol))
	((pld-type string)
	 (signature expression (lambda (x) (or (not x) (number? x))))
	 (optcom optstring))
	((spew "" #f #f)
	 (int "" #f #f)
	 (hid "" #f #f)
	 (* "*" #f #f)))

      '(block
	((name symbol))
	((clock expression)
	 (reset symbol)			;really signal foreign-key
	 (clock-frequency expression)
	 (synchronous-reset? boolean)
	 (optcom optstring))
	())

      '(pin-type
	((name atom))
	()
	((output)
	 (input/output)
	 (wire-and)
	 (wire-or)
	 (clock)
	 (input)
	 (unused)
	 (internal)
	 (library-module)))

      '(pin
	((ptag ptag)
	 (name symbol))
	((pin-type pin-type)
	 (pin-name expression)	 ;'?, number, symbol, or list of same.
	 (output-enable expression)
	 (paired-input expression (lambda (x) (or (boolean? x) (symbol? x))))
	 (defining-block block))
	())

      '(pindex
	((ptag ptag)
	 (pin-name atom))
	((symbol-name symbol)
	 (pin-type pin-type)
	 (name expression)	    ;corresponds to NAME in pin table.
	 (signal-name expression))
	())

      '(pin-basename
	((ptag ptag)
	 (basename symbol))		;name without bounds
	((names expression (lambda (x) (every symbol? x))))
	())

      '(check
	((block block)
	 (order uint))
	((expression expression))
	())

      '(signal-type
	((name symbol))
	()
	((registered)
	 (registered-input)
	 (combinatorial)
	 (hidden)			;used for inputs
	 (macro)))

      '(signal
	((block block)
	 (name symbol))
	((defining-ptag ptag)
	 (signal-type signal-type)
	 (initial-state expression)
	 (next-function expression)
	 (order uint))
	())

      '(ram
	((block block)
	 (name symbol))
	((output expression)
	 (write-enable expression)
	 (address expression)
	 (data expression))
	())

      '(used-in-ptag
	((name symbol)
	 (ptag ptag))
	()
	())

      '(flags
	((block block)
	 (signal-name symbol))
	((flags expression))
	())

      '(u-ptag
	((u-number symbol))
	((ptag ptag))
	())

      '(netlist
	((name symbol)
	 (index atom))
	((sheet atom)
	 (u-pins expression
		 (lambda (x)
		   (and (list? x)
			(every
			 (lambda (elt)
			   (and (pair? elt) (symbol? (car elt))
				(or (integer? (cdr elt))
				    (symbol? (cdr elt)))))
			 x))))
	 (r-pins expression
		 (lambda (x) (and (list? x) (every integer? x)))))
	())

      '(netlist-aggregate
	((name symbol))
	((bounds expression
		 (lambda (b) (if b (and (list? b)
					(integer? (car b))
					(integer? (cadr b)))
				 #t))))
	())

      '(netlist-non-aggregate
	((name symbol))
	()
	())

      '(netlist-inverse
	((ptag ptag)
	 (pin atom (lambda (x) (or (integer? x) (symbol? x)))))
	((net-name symbol)
	 (index atom))
	())

      '(resistor-nets
	((number uint))
	((names expression
		(lambda (x) (and (list? x) (every symbol? x)))))
	())

      '(comment
	((name symbol))
	((optcom optstring))
	())

      )
    boardb))

(define *block* #f)
(defvar *design* 'test)
(define *board* (create-board *design* 'alist-table))

(define (synch:register-block block . args)
  (let* ((clock (if (null? args)
		    (symbol-append block '- 'clock)
		    (car args)))
	 (reset (if (or (null? args) (null? (cdr args)))
		    (symbol-append block '- 'reset-)
		    (cadr args)))
	 (synchronous-reset?
	  (and (pair? reset)
	       (eq? (car reset) 'SYNCHRONOUS)
	       (begin (set! reset (cadr reset))
		      #t)))
	 (clock-frequency
	  (and (> (length args) 1)
	       (some (lambda (x) (and (number? x) x)) args))))
    ((((*board* 'open-table) 'block #t) 'row:insert)
     (list block clock reset clock-frequency synchronous-reset?
	   (comment (symbol-append *design* block))))
    (set! *block* block)))

(define (synch:register-ptag tag ptag . args)
  (define signature (if (null? args) #f (car args)))
  ((ptag-table 'row:insert)
   (list tag ptag signature (comment (symbol-append tag '- ptag)))))

(define ptag-table ((*board* 'open-table) 'ptag #t))
(define ptag->pld (ptag-table 'get 'pld-type))
(define ptag->sig (ptag-table 'get 'signature))

(define pin-table ((*board* 'open-table) 'pin #t))
(define pin-table:row-insert (pin-table 'row:insert))
(define pin-table:row-retrieve (pin-table 'row:retrieve))
(define pin-table:row-update (pin-table 'row:update))

(define pin-basename-table ((*board* 'open-table) 'pin-basename #t))
(define pin-basename:row-retrieve (pin-basename-table 'row:retrieve))
(define pin-basename:update
  (let ((pin-basename:row-update (pin-basename-table 'row:update)))
    (lambda (ptag name pin-type)
      (let* ((bnds (vector-bounds name))
	     (basename (if bnds (car bnds) name))
	     (row (pin-basename:row-retrieve
		   ptag (if bnds (car bnds) name))))
	(if row
	    (or (eq? pin-type
		     ((pin-table 'get 'pin-type) ptag (caaddr row)))
		(synch:warn "Pins with same basename but different types"
			    ptag name (caaddr row))))
	(pin-basename:row-update
	 (list ptag basename
	       (if row (cons name (caddr row)) (list name))))))))

(define ordering:counter 100)
(define signal-table ((*board* 'open-table) 'signal #t))
(define signal-table:row-insert (signal-table 'row:insert))
(define signal-table:row-remove (signal-table 'row:remove))
(define signal-table:row-retrieve (signal-table 'row:retrieve))
(define signal-table:signal-type (signal-table 'get 'signal-type))
(define signal-table:get-init-fun (signal-table 'get 'initial-state))
(define signal-table:get-next-fun
  (let ((signal-type (signal-table 'get 'signal-type))
	(get-next-fun (signal-table 'get 'next-function)))
    (lambda (*block* signal)
      (define sigtype (signal-type *block* signal))
      ;;(print signal 'sigtype sigtype)
      (cond ((not (eq? sigtype 'registered))
	     (synch:error
	      'synch:pre 'of sigtype 'signal: signal)))
      (get-next-fun *block* signal))))

(define (signal-table:for-each-row proc . keys)
  (for-each proc
	    (sort (apply (signal-table 'row:retrieve*) keys)
		  (lambda (row1 row2)
		    (< (list-ref row1 6) (list-ref row2 6))))))

(define flags-table ((*board* 'open-table) 'flags #t))
(define flags:row-update (flags-table 'row:update))
(define flags:row-retrieve (flags-table 'row:retrieve))
(define (synch:signal-flags tag name)
  (let ((row (flags:row-retrieve tag name)))
    (if row (list-ref row 2) '())))

(define comment-table ((*board* 'open-table) 'comment #t))
(define comment-table:row-insert (comment-table 'row:insert))
(define signal->comment
  (let ((getter (comment-table 'get 'optcom)))
    (lambda (spec)
      (getter (if (not (pair? spec))
		  spec
		  (case (length spec)
		    ((1 2 3) (car spec))
		    ((4 5 6) (car (cadddr spec)))
		    (else #f)))))))

(define used-in-ptag ((*board* 'open-table) 'used-in-ptag #t))
(define used-in-ptag:row-update (used-in-ptag 'row:update))
;;(define used-in-ptag:row-remove (used-in-ptag 'row:remove))
(define used-in-ptag:row-retrieve* (used-in-ptag 'row:retrieve*))
(define used-in-ptag:get-ptag (used-in-ptag 'get 'ptag))

(define block-table ((*board* 'open-table) 'block #f))

(define (mark-uses ptag next-function)
  (cond ((symbol? next-function)
	 (used-in-ptag:row-update (list next-function ptag)))
	((and (list? next-function) (pair? next-function))
	 (for-each (lambda (expr) (mark-uses ptag expr))
		   next-function))))

(define (used-by-other-ptags? signal ptag)
  (not (null? (used-in-ptag:row-retrieve*
	       signal (lambda (key) (and (not (eq? key ptag))
					 (not (memq key '(HID *)))))))))
;;(trace used-in-ptag:get-ptag used-by-other-ptags?)

(define constant-table ((*board* 'open-table) 'constants #t))
(define constant-table:insert
  (let ((inserter (constant-table 'row:insert)))
    (lambda args
      (inserter args))))
(define constant-table:row-retrieve (constant-table 'row:retrieve))

(defmacro defconst (name value)
  (cond ((list? name) `(defconst ,(car name) (lambda ,(cdr name) ,value)))
	(else (cond ((not (slib:eval `(defined? ,name))))
		    ((and (symbol? name) (eqv? (slib:eval value)
					       (slib:eval name))))
		    (else (slib:error 'trying-to-defconst name
				      'to-different-value value)))
	      `(begin (define ,name ,value)
		      (constant-table:insert ',name ,name ,(comment name))))))
(defconst *xxxx* #x-f)

;;; Eval constant
(define (ceval thing)
  (cond ((number? thing) thing)
	((not (symbol? thing)) thing)
	((constant-table:row-retrieve thing) =>
	 cadr)
	(else thing)))

(define (for-each-ptag proc)
  ((ptag-table 'for-each-row-in-order)
   (lambda (ptag-row)
     (define str (cadr ptag-row))
     (if (and (not (equal? "" str))
	      (or (char-alphabetic? (string-ref str 0))
		  (char-numeric? (string-ref str 0))))
	 (apply proc ptag-row)))))

(define check-table ((*board* 'open-table) 'check #t))
(define synch:insert-check
  (let ((check-table:row-insert (check-table 'row:insert)))
    (lambda (check)
      (set! ordering:counter (+ ordering:counter 1))
      (check-table:row-insert (list *block* ordering:counter check)))))

(define ram-table ((*board* 'open-table) 'ram #t))
(define ram-table:row-insert (ram-table 'row:insert))
(define ram-table:row-update (ram-table 'row:update))
(define ram-table:row-retrieve (ram-table 'row:retrieve))

;;(SYNCH:DEFINE-PIN
;; ptag name pin-type pin-name output-enable paired-input defining-block)
(define (synch:define-pin ptag . args)
  (cond ((not ptag) (set! ptag '*)))
  (if (not ((ptag-table 'row:retrieve) ptag))
      (synch:register-ptag ptag ""))
  (pin-table:row-insert (cons ptag (append args (list *block*))))
  (apply
   (lambda (name pin-type pin-name output-enable paired-input defining-block)
     (pin-basename:update ptag name pin-type)
     (cond
      ((eq? pin-name '?))
      ((and (or (list? pin-name) (symbol? pin-name))
	    (vector-bounds (or paired-input name)))
       => (lambda (bounds)
	    (let* ((first (cadr bounds))
		   (last (caddr bounds))
		   (range (- last first))
		   (incer (sign range)))
	      (do ((i 0 (+ incer i)))
		  ((positive? (* incer (- i range))))
		(pindex:row-insert
		 (list ptag
		       (if (pair? pin-name)
			   (list-ref pin-name (abs i))
			   (symbol-append pin-name (abs i)))
		       '?
		       pin-type
		       name
		       (refvectorize (or paired-input name) (+ i first))
		       ))))))
      ((or (and (number? pin-name)
		(not (zero? pin-name))
		pin-name)
	   (and (symbol? pin-name)
		(ptag-stripped pin-name)))
       => (lambda (pname)
	    (pindex:row-insert
	     (list ptag pname '? pin-type name name)))))
     (if (comment?)
	 (comment-table:row-insert (list name (comment name)))))
   (append args (list *block*))))

;;(SYNCH:DEFINE-SIGNAL
;; [block] name defining-ptag signal-type initial-state next-function)
(define (synch:define-signal name ptag . args)
  (set! ordering:counter (+ ordering:counter 1))
  (if (and ptag (not ((ptag-table 'row:retrieve) ptag)))
      (synch:register-ptag ptag ""))
  (apply (lambda (block name ptag signal-type initial-state next-function)
	   (if (list? initial-state)
	       (case (car initial-state)
		 ((SYNCHRONOUS)
		  (set! initial-state (cadr initial-state))
		  (synch:add-signal-flag block name 'SYNCHRONOUS-RESET))
		 ((MAKE-ASYN-RAM MAKE-SYNC-RAM MAKE-ROM)
		  (ram-table:row-insert
		   (list block name #f #f #f #f)))))
	   (signal-table:row-insert (list block name ptag signal-type
					  initial-state next-function
					  ordering:counter))
	   (if (comment?)
	       (comment-table:row-insert (list name (comment name)))))
	 *block* name (or ptag '*) args))

(define (synch:define-bus name float-limit . connections)
  (define ptag (ptag-prefix name))
  (cond (ptag (set! name (ptag-stripped name)))
	(else (set! ptag '*)))
  (let* ((nm (car (or (vector-bounds name) (list name))))
	 (nm-drv (symbol-append nm '- 'drv)))
    #+check-turnaround
    (define nm-lst (symbol-append nm '- 'lst))
    (synch:define-signal name #f 'macro #f
			 `(cond ,@connections (else ,(- -1 float-limit))))
    #+check-turnaround
    (synch:define-signal
     nm-lst 'hid 'registered (- -1 float-limit) name)
    (synch:define-signal
     nm-drv 'hid 'registered (- -1 float-limit)
     `(case (count ,@(remove #f (map car connections)))
	((0) (cond ((eqv? -1 ,nm-drv)
		    (synch:warn ',name 'floating ,float-limit 'cycles)
		    -1)
		   ((and (number? ,nm-drv) (negative? ,nm-drv))
		    (+ 1 ,nm-drv))
		   (else ,(- -1 float-limit))))
	((1) (let ((idx (cond ,@(let ((idx -1))
				  (map (lambda (con)
					 (set! idx (+ 1 idx))
					 `(,(car con) ,idx))
				       connections))
			      (else #f))))
	       #+check-turnaround
	       (cond ((eqv? idx ,nm-drv))
		     ((negative? ,nm-drv))
		     ((negative? idx))
		     ((eqv? ,name ,nm-lst))
		     (else (synch:warn ',name 'no-turnaround-time ,nm-drv idx)))
	       idx))
	,@(if (some boolean? (map car connections))
	      `(((2) ,nm-drv))
	      '())
	(else (apply synch:warn ',name 'bus-driving-conflict
		     (list ,@(map (lambda (x) `',(car x)) connections)))
	      (cond ,@(let ((idx -1))
			(map (lambda (con)
			       (set! idx (+ 1 idx))
			       `(,(car con) ,idx))
			     connections))
		    (else #f)))))))

(define (synch:add-signal-flag block signal-name flag)
  (let ((row (flags:row-retrieve block signal-name)))
    (flags:row-update (list block signal-name
			    (if row
				(cons flag (list-ref row 2))
				(list flag))))))

(define (synch:check-flags recognized-flags)
  ((flags-table 'for-each-row-in-order)
   (lambda (row)
     (apply (lambda (block signal flags)
	      (for-each (lambda (flag)
			  (or (memq flag recognized-flags)
			      (synch:warn "unrecognized flag" flag block signal)))
			flags))
	    row))))

(defmacro synch:defcheck (check)
  (tprint 'synch:defcheck check)
  (synch:insert-check check)
  "check")

;;; Synch:pre is a macro that saves typing.

(defmacro synch:pre (signal)
  (tprint 'synch:pre signal)
  (let ((stepper (signal-table:get-next-fun *block* signal)))
    (case stepper
      ((#f) (synch:error "signal not known:" *block* signal))
      ((?) (synch:error "signal not yet set:" *block* signal))
      (else stepper))))

(defmacro synch:defmacro (name expression)
  (define ptag (ptag-prefix name))
  (set! name (ptag-stripped name))
  (tprint 'synch:defmacro ptag name)
  (synch:define-signal name ptag 'macro #f expression)
  (symbol->string name))

(defmacro synch:defshare (name . args)
  (cond ((null? args)
	 (let ((ptag (ptag-prefix name)))
	   (set! name (ptag-stripped name))
	   (tprint 'synch:defshare ptag name)
	   (cond ((signal-table:row-retrieve *block* name)
		  (synch:define-pin ptag name 'input name #f name))
		 (else (synch:error "shared signal not defined:" name)))
	   (symbol->string name)))
	((null? (cdr args))
	 (let* ((pin-nam name)
		(name (car args))
		(ptag (ptag-prefix name)))
	   (set! name (ptag-stripped name))
	   (tprint 'synch:defshare ptag name)
	   (cond ((signal-table:row-retrieve *block* name)
		  (synch:define-pin ptag name 'input (ptag-stripped pin-nam) #f name))
		 (else (synch:error "shared signal not defined:" name)))
	   (symbol->string name)))
	(else (synch:error 'synch:defshare (cons name args)))))

(defmacro synch:definput (pin-nam name expression)
  (define ptag (or (ptag-prefix pin-nam) (ptag-prefix name)))
  (if (eq? ptag (ptag-prefix name)) (set! name (ptag-stripped name)))
  (tprint 'synch:definput ptag name)
  (synch:define-signal name ptag 'hidden #f expression)
  (synch:define-pin ptag name 'input (ptag-stripped pin-nam) #f name)
  (symbol->string name))

(defmacro synch:defreginput (pin-nam fullname expression)
  (let* ((ptag (or (ptag-prefix pin-nam) (ptag-prefix fullname)))
	 (name (ptag-stripped fullname))
	 (fullprename (symbol-append ptag ': 'pre: name)))
    (tprint 'synch:defreginput ptag name)
    `(begin
       (define ,name #f)		;This BAD Hack is spreading!
					;Needed for self-reset test in scm2mach
					;registered inputs clear to #f
       (synch:definput ,(ptag-stripped pin-nam) ,fullprename ,expression)
       (synch:define 0 ,fullname ,name #f #f)
       (synch:set! ,name ,(symbol-append 'pre: name)))))

(defmacro synch:define (pin-nam signal init-state pin oen . opt-next)
  (define ptag (or (ptag-prefix pin-nam) (ptag-prefix signal)))
  (set! opt-next (if (null? opt-next) '? (car opt-next)))
  (tprint 'synch:define ptag signal)
  (cond (ptag
	 (if (eq? ptag (ptag-prefix signal))
	     (set! signal (ptag-stripped signal)))
	 (synch:define-pin ptag signal
			   (case oen
			     ((#f) 'internal)
			     ((#t) 'output)
			     (else 'input/output))
			   (ptag-stripped pin-nam)
			   oen
			   (if oen (or pin signal) pin)
			   ;;pin
			   )))
  (synch:define-signal signal ptag 'registered init-state opt-next)
  (symbol->string signal))

(defmacro synch:defcomb (pin-nam signal pin oen next)
  (define ptag (or (ptag-prefix pin-nam) (ptag-prefix signal)))
  (tprint 'synch:defcomb ptag signal)
  (cond (ptag
	 (if (eq? ptag (ptag-prefix signal))
	     (set! signal (ptag-stripped signal)))
	 (synch:define-pin ptag signal
			   (case oen
			     ((#f) 'internal)
			     ((#t) 'output)
			     (else 'input/output))
			   (ptag-stripped pin-nam)
			   oen
			   (if oen (or pin signal) pin))))
  (synch:define-signal signal ptag 'combinatorial '*xxxx* next)
  (symbol->string signal))

(defmacro synch:set/reset! (signal turn-on turn-off)
  (tprint 'synch:set/reset! signal)
  (synch:insert-check
   `(cond ((and ,turn-on ,turn-off)
	   (synch:warn ',signal 'being-turned-on-and_off_))))
  `(synch:set! ,signal (and (not ,turn-off) (or ,turn-on ,signal))))

(defmacro synch:set! (signal new-state)
  (tprint 'synch:set! signal)
  (let* ((row (signal-table:row-remove *block* signal))
	 (last2 (and row (last row 2))))
    (cond ((not row) (synch:error "signal not known:" signal))
	  ((eq? '? (car last2))
	   (signal-table:row-insert
	    (append (butlast row 2) (cons new-state (cdr last2)))))
	  ((and (pair? (car last2)) (eq? 'compose-sets (caar last2)))
	   (signal-table:row-insert
	    (append (butlast row 2)
		    (cons (append (car last2) `((lambda (,signal) ,new-state)))
			  (cdr last2)))))
	  (else
	   (signal-table:row-insert
	    (append (butlast row 2)
		    (cons `(compose-sets ,signal
					 (lambda (,signal) ,(car last2))
					 (lambda (,signal) ,new-state))
			  (cdr last2))))))
    (symbol->string signal)))

(defmacro synch:declare (signal flag)
  (tprint 'synch:declare signal)
  (synch:add-signal-flag *block* signal flag))

(define synch:step '())
(define synch:status #f)
(defmacro define-synchronous-system (system-name . body)
  (cond ((not (eq? system-name *block*))
	 (slib:error "defining system-name different from *block*?" system-name
		     *block*)))
  (synch:define-signal 'synch:count 'HID 'registered 0 '(+ 1 synch:count))
  ;;(solidify-database *board*)		; make immutable
  ;; bad idea for multi-block designs.
  (sync-database *board*)
  (let ((names '())
	(inits '())
	(macros '())
	(hiddens '())
	(steppers '())
	(block->clock (((*board* 'open-table) 'block #f) 'get 'clock))
	(block->reset (((*board* 'open-table) 'block #f) 'get 'reset))
	(synch:blocks ((((*board* 'open-table) 'block #f) 'get* 'name))))
    (for-each
     (lambda (sig-row)
       (apply
	(lambda (block name defining-ptag signal-type
		       initial-state next-function order)
	  (case signal-type
	    ((registered combinatorial macro)
	     (mark-uses defining-ptag next-function)
	     (and (not (eqv? '*xxxx* initial-state))
		  (mark-uses defining-ptag initial-state))))
	  (case signal-type
	    ((registered) (if (eq? '? next-function)
			      (synch:error "signal not set:" sig-row))
	     (cond
	      ((and (not (pair? initial-state)) (vector-bounds name)) =>
	       (lambda (bounds)
		 (set! next-function
		       `(chkbnd ',name
				,(+ -1 (ash 1 (+ 1 (- (cadr bounds)
						      (caddr bounds)))))
				,next-function)))))
	     (set! names (cons name names))
	     (set! inits (cons initial-state inits))
	     (set! steppers (cons next-function steppers)))
	    ((hidden) (set! hiddens (cons (list name next-function) hiddens)))
	    ((macro combinatorial)
	     (set! macros (cons (list name next-function) macros)))
	    (else (synch:error "unknown signal-type" sig-row))))
	sig-row))
     (sort ((signal-table 'row:retrieve*) *block*)
	   (lambda (row1 row2) (> (list-ref row1 6) (list-ref row2 6)))))
    (set! synch:step
	  (slib:eval
	   `(lambda ()
;;;	      ,@(map (lambda (sys-name)
;;;		       (define time-name (symbol-append sys-name ':time))
;;;		       `(if ,time-name
;;;			    (set! ,time-name (+ ,time-name synch:inc))))
;;;		     synch:blocks)
	      (or?
	       ,@(map (lambda (sys-name)
			;;(define time-name (symbol-append sys-name ':time))
			(define clock-expr (block->clock sys-name))
			(define clock-state (symbol-append sys-name ':clock))
			(define reset-name (block->reset sys-name))
			(define state-name (symbol-append sys-name ':state))
			`(cond (,(if (negated? reset-name)
				     `(not ,reset-name)
				     reset-name)
				(set! ,state-name #f)
				#f)
			       ((not ,state-name)
				(set! ,state-name
				      (slib:eval
				       (cons 'list
					     ,(symbol-append
					       sys-name ':state-initializers))))
				;;(if (number? ,clock-expr) (set! ,time-name 0))
				#f)
;;;			       ((and ,time-name (> ,time-name ,clock-expr))
;;;				(set! ,time-name (- ,time-name ,clock-expr))
;;;				(set! ,state-name
;;;				      (@apply
;;;				       ,(symbol-append sys-name ':state-step)
;;;				       ,state-name))
;;;				#t)
;;;			       (,time-name #t)
			       ((positive? ,(symbol-append sys-name ':queue))
				(set! ,(symbol-append sys-name ':queue)
				      (+ -1 ,(symbol-append sys-name ':queue)))
				(set! ,state-name
				      (@apply
				       ,(symbol-append sys-name ':state-step)
				       ,state-name))
				#t)
			       (else #f)))
		      synch:blocks)))))	;(reverse )
    (set! synch:status
	  (slib:eval
	   `(lambda ()
	      (synch:prin-messages!)
	      ;;(display "...")
	      (if synch:stop (printf "STOPPED!"))
	      ,@(map (lambda (block)
		       (let ((reset-name (block->reset block)))
			 `(apply printf " %s: %s q=%d r=%d p=%%d"
				 ',(string-upcase (symbol->string block))
				 (if ,(if (negated? reset-name) reset-name
					  `(not ,reset-name))
				     "run"
				     "rst")
				 ,(symbol-append block ':queue)
				 (reverse
				  (last (or ,(symbol-append block ':state)
					    (list 0 0))
					2)))))
		     synch:blocks)	;(reverse )
	      (newline))))
    `(begin
       (define ,(symbol-append system-name ':clock)
	 ',(block->clock system-name))
       ;;(define ,(symbol-append system-name ':time) #f)
       (define ,(symbol-append system-name ':queue) 0)
       (define (,(symbol-append system-name '-sim) num)
	 (set! ,(symbol-append system-name ':queue)
	       (max num ,(symbol-append system-name ':queue)))
	 (kb:announce-polling 'break)
	 (simulate!))
       (define ,(symbol-append system-name ':state) #f)
       (define ,(symbol-append system-name ':state-initializers) ',inits)
       (defmacro ,system-name (name)
	 (if (and (list? name) (pair? name) (eq? 'quote (car name)))
	     (set! name (cadr name)))
	 (let ((vals (symbol-append ',system-name ':state))
	       (idx (position name ',names)))
	   (if (not idx)
	       (slib:error name 'not-registered-in vals)
	       `(if ,vals
		    (list-ref ,vals ,idx)
		    (slib:eval
		     (list-ref
		      ,',(symbol-append system-name ':state-initializers)
		      ,idx))))))
       (define (,(symbol-append system-name ':state-step) ,@names)
	 (fluid-let ((*block* ',*block*))
	   (let* ,hiddens
	     (let* ,macros
	       ,@(map caddr
		      (sort ((check-table 'row:retrieve*) *block*)
			    (lambda (row1 row2) (> (cadr row1) (cadr row2)))))
	       ,@body
	       (let ((new-state (list ,@steppers)))
		 (update-deferred-sets!)
		 new-state))))))))

(define simulating? #f)
(define (simulate!)
  (cond (simulating?)
	(synch:stop (synch:status) (kb:break))
	((kb:break-on-hit))
	(else (set! simulating? #t)
	      (let ((more-to-do? (synch:step)))
		(cond ((and (not (null? synch:messages))
			    (not print-timing))
		       (synch:prin-messages!)))	;messages from non-printing
		(set! simulating? #f)
		(if more-to-do? (simulate!))))))

(define synch:stop #f)
(define (synch:reset) (set! synch:stop #f))

;;; Print out as many warnings and errors as possible before stopping.
(require 'object->string)
(defvar print-timing #t)
(defvar print-chat #t)
(define synch:chattel '())
(define synch:messages '())
(defvar talkative #t)
(define (tprint . args)
  ;;(if talkative (apply print args))
  #f)

(defvar report slib:error)

(define (synch:chat msg)
  (set! synch:chattel (cons msg synch:chattel))
  #t)

(define (printf:format? str)
  (cond ((not (string? str)) #f)
	((< 100 (string-length str)) #f)
	(else (memv #\% (string->list str)))))

(define (synch:shout grade args)
  (cond (simulating?
	 (set! synch:messages
	       (cons
		(call-with-output-string
		    (lambda (cep)
		      (fprintf cep "%S: " grade)
		      (cond
		       ((printf:format? (car args))
			(apply fprintf cep args))
		       ((not (null? args))
			(display (car args) cep)
			(for-each (lambda (x)
				    (display #\space cep)
				    (if (number? x)
					(display
					 (string-append "#x" (number->string x 16))
					 cep)
					(write x cep)))
				  (cdr args))))))
		synch:messages)))
	((equal? "ERROR" grade)
	 (apply slib:error args))
	(else (apply slib:warn args))))

(define (synch:info . args)
  (synch:shout "INFO" args))

(define synch:log-port #f)
(define synch:nlog 0)
(define synch:triggers '())
(define (synch:open-log filename)
  (set! synch:log-port filename)
  (and (file-exists? filename)
       (call-with-input-file filename
	 (lambda (port)
	   (do ((trig (read port) (read port))
		(triggers '() (cons trig triggers)))
	       ((eof-object? trig)
		(set! synch:triggers (reverse triggers))))))))

(define (synch:do-msgs synch:count)
  (cond ((and (not (null? synch:messages))
	      synch:log-port
	      (< synch:nlog 50)
	      (string? (car synch:messages))
	      (< 4 (string-length (car synch:messages)))
	      (memv (string-ref (car synch:messages) 0) '(#\W #\E)))
	 (if (string? synch:log-port)
	     (set! synch:log-port (open-output-file synch:log-port)))
	 (fprintf synch:log-port "%d ;;%s\n" synch:count
		  (car synch:messages))
	 (force-output synch:log-port)
	 (set! synch:nlog (+ 1 synch:nlog))))
  (let loop ()
    (cond ((and (not (null? synch:triggers))
		(<= (car synch:triggers) synch:count))
	   (if (null? synch:messages)
	       (set! synch:messages (cons (car synch:triggers) synch:messages)))
	   (set! synch:triggers (cdr synch:triggers))
	   (cond ((not print-timing) (set! print-timing 50))
		 ((number? print-timing)
		  (set! print-timing (max 50 print-timing))))
	   (loop))))
  (not (null? synch:messages)))

(define (synch:warn . args)
  (synch:shout "WARN" args)
  (cond ((not print-timing) (set! print-timing 50))
	((number? print-timing)
	 (set! print-timing (max 50 print-timing)))))

(define (synch:error . args)
  (synch:shout "ERROR" args)
  (cond ((not print-timing) (set! print-timing 50))
	((number? print-timing)
	 (set! print-timing (max 50 print-timing))))
  (set! synch:stop #t))

(define (synch:prin-header names)
  (cond ((and (list? names) (pair? names) (eq? (car names) 'quote))
	 (set! names (cadr names))))
  (let* ((ss (map (lambda (i)
		    (cond ((string? i) (string-upcase! i))
			  ((char? i) "")
			  (else (object->string i))))
		  names))
	 (sl (map string-length ss))
	 (ml (apply max sl))
	 (ll (map (lambda (s l)
		    (string->list
		     (if (zero? l) (make-string ml #\nul)
			 (string-append s (make-string (- ml l) #\space)))))
		  ss sl))
	 (rl (apply map list ll)))
    (for-each printf
	      (map (lambda (l)
		     (string-append
		      (apply
		       string-append
		       " "
		       (if (char=? #\nul (car l)) "" (string (car l)))
		       (map (lambda (c)
			      (string-append
			       "  " (if (char=? #\nul c) "" (string c))))
			    (cdr l)))
		      "\\n"))
		   rl))
    (newline)))

(define (synch:prin1 x)
  (case x
    ((#t) (display "  |"))
    ((#f) (display " | "))
    (else
     (display #\space)
     (cond ((char? x) (display x))
	   ((number? x)
	    (cond ((eqv? *xxxx* x) (display "??"))
		  ((negative? x) (display "  "))
		  (else
		   (let ((s (number->string x 16))
			 (d 2))
		     (do ((l (string-length s) (+ 1 l)))
			 ((>= l d) (display s))
		       (display #\0))))))
	   ((string? x) (display x))
	   (else (write x))))))

(defconst synch:prindepth 100)
(define synch:prindex 0)
(define synch:countsave (make-array '#() synch:prindepth))
(define synch:prinsave (make-array '#() synch:prindepth))
(define synch:chatsave (make-array '#() synch:prindepth))
(define synch:prin-count -1)
(define synch:last-printed-count 0)

(define (synch:prline1 args chat count idx names)
  (cond ((zero? (remainder synch:prin-count 50))
	 (printf "%4d\\n"  count)
	 (synch:prin-header names)))
  (set! synch:prin-count (+ 1 synch:prin-count))
  (array-set! synch:prinsave #f idx)
  (array-set! synch:chatsave #f idx)
  (array-set! synch:countsave #f idx)
  (if args (for-each synch:prin1 args))
  (if chat (for-each (lambda (m) (display #\space) (display m)) chat))
  (newline))

(define (synch:print-pre-trigger synch:count names)
  (define pcount? #f)
  (do ((idx (modulo (+ 1 synch:prindex) synch:prindepth)
	    (modulo (+ 1 idx) synch:prindepth))
       (lidx #f idx))
      ((eqv? lidx synch:prindex))
    (let ((args (array-ref synch:prinsave idx))
	  (chat (array-ref synch:chatsave idx))
	  (count (array-ref synch:countsave idx)))
      (cond (pcount?)
	    ((not count))
	    ((eqv? synch:last-printed-count (+ -1 synch:count))
	     (set! pcount? #t))
	    (else (printf "**************** %d - %d ****************\n"
			  synch:last-printed-count count)
		  (set! synch:prin-count (+ 1 synch:prin-count))
		  (set! pcount? #t)))
      (cond ((and args (or (not (null? args))
			   (not (null? chat))))
	     (synch:prline1 args chat count idx names))))))

(define (synch:prin synch:count names args)
  (define new?
    (or (not (equal? args (array-ref synch:prinsave synch:prindex)))
	(not (null? synch:chattel))))
  (cond ((synch:do-msgs synch:count)
	 (cond ((number? print-timing)
		(synch:print-pre-trigger synch:count names)))
	 (synch:prin-messages!)))	;clears synch:messages
  (cond
   (print-timing
    (cond ((not (number? print-timing)))
	  ((zero? print-timing) (set! print-timing #f))
	  (else (set! print-timing (+ -1 print-timing))))
;;;    (cond ((not (eqv? synch:last-printed-count (+ -1 synch:count)))
;;;	      (synch:status) (set! synch:prin-count (+ 1 synch:prin-count))))
    (cond (new?
	   (synch:prline1 args synch:chattel synch:count synch:prindex names)))
    (set! synch:last-printed-count synch:count))
   ((and print-chat (not (null? synch:chattel)))
    (printf "%5.5d" synch:count)
    (for-each (lambda (m) (display #\space) (display m)) synch:chattel)
    (newline)))
  (cond (new?
	 (set! synch:prindex (modulo (+ 1 synch:prindex) synch:prindepth))
	 (array-set! synch:prinsave args synch:prindex)
	 (array-set! synch:chatsave synch:chattel synch:prindex)
	 (array-set! synch:countsave synch:count synch:prindex)))
  (set! synch:chattel '()))

(define (synch:prin-messages!)
  (for-each (lambda (line) (write-line line (current-error-port)))
	    synch:messages)
  (set! synch:prin-count (+ (length synch:messages) synch:prin-count))
  (set! synch:messages '()))

(defmacro synch:print args
  (define names '())
  (define argl '())
  (do ((args args (cdr args)))
      ((null? args))
    (cond ((string? (car args))
	   (set! names (cons (car args) names))
	   (set! argl (cons (cadr args) argl))
	   (set! args (cdr args)))
	  ((char? (car args))
	   (set! names (cons (car args) names))
	   (set! argl (cons (car args) argl)))
	  ((and (symbol? (car args)) (vector-bounds (car args)))
	   => (lambda (bnds)
		(define wid (+ -1 (bounds->width bnds)))
		(cond
		 ((<= 0 (modulo wid 8) 3)
		  (set! names (cons (symbol->string (car bnds)) names))
		  (set! argl (cons `(,(symbol-append 'nibble (quotient wid 8))
				     ,(car args))
				   argl))
		  (set! wid (+ -4 wid))))
		(do ((wid wid (+ -8 wid)))
		    ((not (positive? wid)))
		  (set! names (cons (symbol->string (car bnds)) names))
		  (set! argl (cons `(,(symbol-append 'byte (quotient wid 8))
				     ,(car args))
				   argl)))))
	  (else
	   (set! names (cons (object->string (car args)) names))
	   (set! argl (cons (car args) argl)))))
  `(synch:prin synch:count ',(reverse names) (list ,@(reverse argl))))

;   ================================================================

(define pindex ((*board* 'open-table) 'pindex #t))
(define pindex:row-insert (pindex 'row:insert))
(define pindex:row-retrieve (pindex 'row:retrieve))

(define (expand-range pair)
  (if (list? pair) pair
      (do ((l '() (cons j l))
	   (inc (if (> (cdr pair) (car pair)) -1 +1))
	   (j (cdr pair) (+ inc j)))
	  ((= j (car pair)) (cons j l)))))

(define (db:read-pins ptag name)
  (define (insert pnum pnam ptyp)
    (let ((s (assq ptyp '((out . output)
			  (i/o . input/output)
			  (hiZ . input/output)
			  ;;(wire-and)
			  ;;(wire-or)
			  (clk . clock)
			  (in . input)
			  (v . unused)
			  ;;(internal)
			  ))))
      (if s
	  (pindex:row-insert (list ptag pnum pnam (cdr s) #f #f))
	  (synch:warn 'db:read-pins "unknown pin type" ptyp))))
  (call-with-input-file (string-append name ".pin")
    (lambda (iport)
      (let ((shape (read iport)))
	(cond
	 ((not (list? shape)) (slib:error "not valid format" shape))
	 ((memq (car shape) '(pqfp plcc lcc soj))
	  (set-car! (cdr shape) (expand-range (cadr shape)))
	  (do ((rows (cadr shape) (cdr rows)))
	      ((null? rows))
	    (let* ((pnum (read iport))
		   (pnam (read iport))
		   (ptyp (read iport)))
	      (cond ((not (equal? pnum (car rows)))
		     (slib:error "pins out of sequence"
				 (car rows) "~=" pnum)))
	      (insert pnum pnam ptyp))))
	 ((memq (car shape) '(bga))
	  (let loop ((c (peek-char iport)))
	    (cond ((eof-object? c))
		  ((char-whitespace? c)
		   (read-char iport)
		   (loop (peek-char iport)))
		  (else
		   (let* ((pnum (read iport))
			  (pnam (read iport))
			  (ptyp (read iport)))
		     (insert pnum pnam ptyp)
		     (loop (peek-char iport)))))))
	 (else (slib:error "unknown format" shape)))))))

(define ptag:row-retrieve (ptag-table 'row:retrieve))
(define u-ptag ((*board* 'open-table) 'u-ptag #t))
(define u-ptag:row-retrieve (u-ptag 'row:retrieve))

(define db:register-u-ptag
  (let ((u-ptag:row-insert (u-ptag 'row:insert)))
    (lambda (u-number ptag)
      (if (ptag:row-retrieve ptag)
	  (u-ptag:row-insert (list u-number ptag))
	  (let* ((chip (object->string ptag))
		 (ptag (symbol-append chip '- u-number)))
	    (synch:register-ptag ptag "")
	    (u-ptag:row-insert (list u-number ptag)))))))

;; Reads a file containing a list of lists (U-NUMBER PTAG-NAME)
(define (db:read-u-ptag name)
  (for-each (lambda (row)
	      (apply db:register-u-ptag row))
	    (call-with-input-file name read)))

(define netlist ((*board* 'open-table) 'netlist #t))
(define netlist:row-retrieve (netlist 'row:retrieve))
(define netlist-inverse ((*board* 'open-table) 'netlist-inverse #t))
(define netlist-inverse:row-retrieve (netlist-inverse 'row:retrieve))
(define netlist-aggregate ((*board* 'open-table) 'netlist-aggregate #t))
(define netlist-aggregate:row-retrieve (netlist-aggregate 'row:retrieve))
(define netlist-non-aggregate
  ((*board* 'open-table) 'netlist-non-aggregate #t))
(define netlist-non-aggregate:row-retrieve
  (netlist-non-aggregate 'row:retrieve))
(define resistor-nets ((*board* 'open-table) 'resistor-nets #t))
(define resistor-nets:row-retrieve (resistor-nets 'row:retrieve))

(define (db:read-net name)
  (require 'scanf)
  (call-with-input-file (if (has-suffix? name ".net")
			    name
			    (string-append name ".net"))
    (lambda (port)
      (define (read1)
	(let ((line (read-line port)))
	  (cond ((eof-object? line) #f)
		((string=? line "$END") #f)
		((zero? (string-length line)) #f)
		((char=? #\, (string-ref line (- (string-length line) 1)))
		 (string-set! line (- (string-length line) 1) #\space)
		 (cons line (read1)))
		(else (list line)))))
      ;; Returns list of (NAME INDEX SHEET)
      ;; NAME a string, INDEX and SHEET either #f or integers.
      (define (parse-name str)
	(let* ((n (string-length str))
	       (i/ (string-reverse-index str #\/))
	       (sheet (and i/ (string->number (substring str (+ i/ 1) n))))
	       (iend (or i/ n)))
	  (if sheet			;otherwise unnamed net
	      (let loop ((i iend))
		(cond ((zero? i)
		       (list (substring str 0 iend) #f sheet))
		      ((char-numeric? (string-ref str (- i 1)))
		       (loop (- i 1)))
		      ((> iend i)
		       (list (substring str 0 i)
			     (string->number (substring str i iend))
			     sheet))
		      (else
		       (list (substring str 0 i) #f sheet))))
	      (list str #f #f))))
      ;; Returns list (SIGNAL-NAME ((PTAG . PIN) ...))
      ;; where SIGNAL-NAME and PTAG are symbols, PIN an integer.
      (define (read-net-entry)
	(let ((s (read1)))
	  (and s
	       (call-with-input-string (apply string-append s)
		 (lambda (stp)
		   (let ((net-name #f))
		     (fscanf stp " %[^;]" net-name)
		     (let ((parsed (parse-name net-name)))
		       (or (eq? #\; (read-char stp))
			   (synch:warn "strange netlist lines" s))
		       (let loop ((u-pins '())
				  (r-pins '()))
			 (let ((c (peek-char stp)))
			   (cond
			    ((eof-object? c)
			     (netlist-insert parsed u-pins r-pins)
			     #t)
			    ((char-whitespace? c)
			     (read-char stp)
			     (loop u-pins r-pins))
			    ((char=? c #\U)
			     (let ((u-number #f)
				   (pin #f))
			       (fscanf stp "%[^.].%s" u-number pin)
			       (let ((u-sym (string->symbol
					     (string-downcase u-number)))
				     (pin (and pin
					       (or (string->number pin)
						   (string->symbol
						    (string-downcase pin))))))
				 (let try ((i 0))
				   (cond ((u-ptag:row-retrieve u-sym) =>
					  (lambda (row)
					    (loop
					     (cons (cons (cadr row) pin)
						   u-pins)
					     r-pins)))
					 ((> i 2)
					  (slib:error 'db:read-net
						      "can't register" u-sym))
					 (else
					  (db:register-u-ptag u-sym 'unknown)
					  (try (+ i 1))))))))
			    ((char=? c #\R)
			     (let ((r-number #f))
			       (fscanf stp "R%d.%*d" r-number)
			       (loop u-pins
				     (if r-number
					 (cons r-number r-pins)
					 r-pins))))
			    (else
			     (fscanf stp " %*s")
			     (loop u-pins r-pins))))))))))))
      (define netlist-insert
	(let ((netlist:row-insert (netlist 'row:insert))
	      (netlist-inverse:row-insert (netlist-inverse 'row:insert))
	      (resistor-nets:row-update (resistor-nets 'row:update)))
	  (lambda (parsed u-pins r-pins)
	    (let ((name (string->symbol (string-downcase (car parsed))))
		  (index (cadr parsed))
		  (sheet (caddr parsed)))
	      (netlist:row-insert
	       (list name (or index -1) sheet u-pins r-pins))
	      (netlist-agg-update name index)
	      (for-each (lambda (pin)
			  (netlist-inverse:row-insert
			   (list (car pin) (cdr pin) name index)))
			u-pins)
	      (for-each (lambda (r-num)
			  (let ((row (resistor-nets:row-retrieve r-num)))
			    (resistor-nets:row-update
			     (list r-num
				   (cons name
					 (if row (cadr row) '()))))))
			r-pins)))))
      (define netlist-agg-update
	(let ((agg-update (netlist-aggregate 'row:update))
	      (non-agg-insert (netlist-non-aggregate 'row:insert)))
	  (lambda (name index)
	    (if index
		(let ((agg-row (netlist-aggregate:row-retrieve name)))
		  (agg-update
		   (if agg-row
		       (let ((bounds (cadr agg-row)))
			 (list name
			       (list (min index (car bounds))
				     (max index (cadr bounds)))))
		       (list name (if index (list index index) #f)))))
		(non-agg-insert (list name))))))

      (let search ((line (read-line port)))
	(cond ((eof-object? line)
	       (slib:error 'read-net "end of file" fname))
	      ((string=? line "$NETS")
	       (let loop ()
		 (if (read-net-entry) (loop))))
	      (else (search (read-line port))))))))

(define (db:print-power-connections)
  (define (power? name)
    (or (memq name '(vcc gnd))
	(let* ((str (symbol->string name))
	       (n (string-length str)))
	  (and (> n 2)
	       (memv (string-ref str 0) '(#\+ #\-))
	       (char=? (string-ref str (- n 1)) #\v)))))
  ;; Returns list of power or ground net names.
  (define (net-connects row)
    (apply
     (lambda (name index sheet u-pins r-pins)
       (let loop ((r-nets (apply append
				 (map (lambda (r)
					(cadr (resistor-nets:row-retrieve r)))
				      r-pins)))
		  (p-nets '()))
	 (cond ((null? r-nets)
		p-nets)
	       ((memq (car r-nets) p-nets)
		(loop (cdr r-nets)
		      p-nets))
	       ((power? (car r-nets))
		(loop (cdr r-nets)
		      (cons (car r-nets) p-nets)))
	       (else
		(loop (cdr r-nets)
		      p-nets)))))
     row))
  (define (print-net name bounds)
    (let ((connects '()))
      ((netlist 'for-each-row-in-order)
       (lambda (row)
	 (set! connects (union (net-connects row) connects)))
       name #f)
      (printf "%10s:"
	      (cond ((not bounds)
		     name)
		    ((= (car bounds) (cadr bounds))
		     (symbol-append name (car bounds)))
		    (else
		     (sprintf #f "%s[%d:%d]" name
			      (car bounds) (cadr bounds)))))
      (for-each
       (lambda (x) (printf " %7s" x))
       connects))
    (newline))

  ((netlist-aggregate 'for-each-row-in-order)
   (lambda (row)
     (apply print-net row)))

  ((netlist-non-aggregate 'for-each-row-in-order)
   (lambda (row)
     (print-net (car row) #f))))


(define (merge-bounds lst alst)
  (let ((pair (assq (car lst) alst)))
    (cond ((not pair) (cons lst alst))
	  ((null? (cdr pair))
	   ;;(report 'multi-declaration-of pair)
	   alst)
	  ((null? (cdr lst))
	   ;;(report 'multiple-declaration-of lst)
	   alst)
	  ((not (cadr pair))
	   (report 'replicated 'array? lst))
	  ((< (cadr pair) (caddr pair))
	   (set-car! (list-tail pair 1) (min (cadr pair) (cadr lst)))
	   (set-car! (list-tail pair 2) (max (caddr pair) (caddr lst)))
	   (set-car! (list-tail pair 3) (union (cadddr pair) (cadddr lst)))
	   alst)
	  (else
	   (set-car! (list-tail pair 1) (max (cadr pair) (cadr lst)))
	   (set-car! (list-tail pair 2) (min (caddr pair) (caddr lst)))
	   (set-car! (list-tail pair 3) (union (cadddr lst) (cadddr pair)))
	   alst))))

(define (get-bounds name)
  (let ((vb (vector-bounds name)))
    (if vb (append vb (list (list name))) (list name))))

(define (call-with-classified-pins tag block proc)
  (define outputs '())
  (define inouts '())
  (define inputs '())
  (define registers '())
  (define lift-bus? (provided? 'lift-bus))
  ((pin-table 'for-each-row-in-order)
   (lambda (pin-row)
     (apply
      (lambda (ptag name pin-type pin-number output-enable paired-input defining-block)
	(let ((lst (get-bounds name))
	      (typ (and block (signal-table:signal-type block name))))
	  (cond
	   ((or (not block) typ)
	    (case pin-type
	      ((OUTPUT)
	       (set! outputs (merge-bounds lst outputs))
	       (cond ((memq typ '(REGISTERED REGISTERED-INPUT))
		      (set! registers (merge-bounds lst registers)))))
	      ((INPUT CLOCK)
	       (set! inputs (merge-bounds lst inputs)))
	      ((INPUT/OUTPUT)
	       (let ((flags (synch:signal-flags defining-block name)))
		 ;;(@print ptag name pin-type pin-number output-enable paired-input defining-block)
		 (cond
		  ((or lift-bus? (memq 'LIFT-BUS flags))
		   (if (symbol? output-enable)
		       (set! outputs (merge-bounds (get-bounds output-enable) outputs))
		       (print 'output-enable 'not 'symbol? output-enable))
		   (cond ((not (and (eq? paired-input name)
				    (not (boolean? output-enable))))
			  (set! outputs (merge-bounds lst outputs)))))
		  (else
		   (set! inouts (merge-bounds (get-bounds paired-input) inouts)))))
	       (set! registers (merge-bounds lst registers)))
	      ((INTERNAL)
	       (let* ((shape (array-bounds (or block defining-block) name))
		      (newb (if (pair? shape)
				(if (null? (cdr lst))
				    (append lst '(#f #f) shape)
				    (append lst shape))
				lst)))
		 (cond (tag
			(cond ((used-by-other-ptags? name tag)
			       (set! outputs (merge-bounds newb outputs))))
			(set! registers (merge-bounds newb registers)))
		       (else
			(cond ((used-in-ptag:get-ptag name 'spew)
			       (set! registers (merge-bounds newb registers))
			       (set! outputs (merge-bounds newb outputs))))))))
	      ((WIRE-AND WIRE-OR)
	       (report "found wire" pin-row))))
	   ((used-in-ptag:get-ptag name tag)
	    (set! inputs (merge-bounds lst inputs))))))
      pin-row))
   tag)
  ;;(@pprint outputs inouts inputs registers)
  (proc outputs inouts inputs registers))

;; Given a list of vectorized signal names SIGS, return an expression
;; combining all into one vectorized signal.
(define (bus-expression sigs)
  (let* ((bnds (sort (map get-bounds sigs)
		     (lambda (b1 b2)
		       (< (cadr b1) (cadr b2)))))
	 (bnd1 (car bnds)))
    (define (mult a b)
      (cond ((eqv? 1 a) b)
	    ((eqv? 1 b) a)
	    (else `(* ,a ,b))))
    (define (add a b)
      (cond ((eqv? 0 a) b)
	    ((eqv? 0 b) a)
	    (else `(+ ,a ,b))))
    (let recur ((cofs (map (lambda (bnd) (car (cadddr bnd)))
			   bnds))
		(pows (if (< (cadr bnd1) (caddr bnd1))
			  (map cadr bnds)
			  (map caddr bnds)))
		(pow 0))
      (if (null? cofs)
	  0
	  (mult (expt 2 (- (car pows) pow))
		(add (car cofs)
		     (recur (cdr cofs)
			    (cdr pows)
			    (car pows))))))))


;; BOUNDS may be an integer, a 2-list of (LOWER-BOUND UPPER-BOUND), or #F.
;; Output is a list of lists : ((PTAG SIGNAL-NAME) ...)
(define (db:netlist-connects net-name bounds)
  (define (union s1 s2)
    (cond ((null? s1) s2)
	  ((null? s2) s1)
	  ((member (car s1) s2) (union (cdr s1) s2))
	  (else (union (cdr s1) (cons (car s1) s2)))))
  (if (list? bounds)
      (let ((imin (apply min bounds))
	    (imax (apply max bounds)))
	(let loop ((i imin)
		   (ret '()))
	  (if (> i imax)
	      ret
	      (loop (+ i 1)
		    (union (db:netlist-connects net-name i) ret)))))
      (let ((row (netlist:row-retrieve net-name (or bounds -1))))
	(and row
	     (apply
	      (lambda (name index sheet u-pins r-pins)
		(let loop ((us u-pins)
			   (ret '()))
		  (if (null? us)
		      ret
		      (loop
		       (cdr us)
		       (let ((pindex-row (pindex:row-retrieve (caar us) (cdar us))))
			 (if pindex-row
			     (cons (list (caar us) (list-ref pindex-row 4)) ret)
			     ret))))))
	      row)))))

;;(trace synch:prin-header synch:print-pre-trigger simulate! synch:prline1 synch:do-msgs synch:prin synch:prin-messages!)(set! *qp-width* 333)

;;(trace synch:define-signal synch:define-pin synch:register-block)
;;(trace-all "../synch/simsynch.scm" "../synch/logic.scm")(untrace tprint)(set! *qp-width* 333)
