;;;; "uart.scm" SimSynch example UART design.	-*-scheme-*-
;;; Copyright (C) 2000 Aubrey Jaffer

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.

;;;================================================================
;;;				 UART
;;;================================================================

;; Default *design* if not already set.
(defvar *design* 'uart)
(require 'simsynch)

#;UART Design.
(synch:register-ptag 'uart "16550")

;;;================================================================
;;;			Modem Signal Interface
;;;================================================================

;;; Modem-status-register inputs
(defvar uart:cts #t)
(defvar uart:dsr #t)
(defvar uart:ri  #t)
(defvar uart:dcd #t)

;;; Modem-control-register outputs
(defvar uart:dtr #f)
(defvar uart:rts #f)
(defvar uart:out1 #f)
(defvar uart:out2 #f)

;;;================================================================
;;;			 uProcessor interface
;;;================================================================

#;This block contains the 8-bit register interface.  To use:
#;[1] set uart:address to select desired register.
#;[2] set uart:write-data to desired contents for write.
#;[3] set uart:read-en #t for read; uart:write-en #t for write.
#;[4] simulate one UIF clock with (uif-sim 1).
#;[5] clear uart:read-en or uart:write-en.
#;[6] read register contents from (uart udata-out[7:0]).
#;[7] simulate one UIF clock with (uif-sim 1).
(synch:register-block 'uif 'uif-strobe)
(synch:definput uart: uif-reset- #t)
(defvar uif-reset- #t)
(synch:definput uart: uif-strobe #t)
(defvar uif-strobe #t)
(defvar uart:address 0)
(defvar uart:write-data 0)
(defvar uart:udata-out *xxxx*)
;;(synch:defmacro uart:uart-clock (or uif-read uif-write))

#;Reads from register addressed by uaddr[2:0] (onto udata-out[7:0]).
(defvar UART:read-en #f)
(synch:defmacro UART:uif-read UART:read-en)
#;Writes data from di[7:0] to register addressed by uaddr[2:0].
(defvar UART:write-en #f)
(synch:defmacro UART:uif-write UART:write-en)
#;UART register address.
(synch:definput UART: uaddr[2:0] uart:address)

#;Divisor Latch Access Bit.
(synch:defmacro UART:dlab (logbit? 7 lcr[7:0]))

#;UART register data.
(synch:definput uart: udata[7:0] uart:write-data)
(synch:define UART: udata-out[7:0] 0 udata[7:0] (and uart:read-en)
	      (case uaddr[2:0]
		((0) (if dlab dll[7:0] rbr[7:0]))
		((1) (if dlab dlm[7:0] (logand #x0f ier[7:0])))
		((2) (logand #xcf iir[7:0]))
		((3) lcr[7:0])
		((4) mcr[7:0])
		((5) lsr[7:0])
		((6) (BAUDx16 msr[7:0]))
		((7) scr[7:0])))

#;Transmit holding register.
(synch:define UART: tbr[7:0] #x00 #f #f)
#;Interrupt enable register.
(synch:define UART: ier[7:0] #x00 #f #f)
#;FIFO control register.
(synch:define UART: fcr[7:0] #x01 #f #f)
#;Line control register.
(synch:define UART: lcr[7:0] #x00 #f #f)
#;MODEM control register.
(synch:define UART: mcr[7:0] #x00 #f #f)
#;Scratch register.
(synch:define UART: scr[7:0] #x00 #f #f)

(synch:set! tbr[7:0] (if (and (= uaddr[2:0] 0) uif-write) udata[7:0] tbr[7:0]))
(synch:set! ier[7:0] (if (and (= uaddr[2:0] 1) uif-write) udata[7:0] ier[7:0]))
(synch:set! fcr[7:0] (if (and (= uaddr[2:0] 2) uif-write) udata[7:0] fcr[7:0]))
(synch:set! lcr[7:0] (if (and (= uaddr[2:0] 3) uif-write) udata[7:0] lcr[7:0]))
(synch:set! mcr[7:0] (if (and (= uaddr[2:0] 4) uif-write) udata[7:0] mcr[7:0]))
(synch:set! scr[7:0] (if (and (= uaddr[2:0] 7) uif-write) udata[7:0] scr[7:0]))

(synch:defcheck
 (cond ((and (= uaddr[2:0] 4) uif-write)
	(set! uart:dtr (logtest? 0 udata[7:0]))
	(set! uart:rts (logtest? 1 udata[7:0]))
	(set! uart:out1 (logtest? 2 udata[7:0]))
	(set! uart:out2 (logtest? 3 udata[7:0])))))

#;Divisor register, least significant bits.
(synch:define UART: dll[7:0] #x00 #f #f)
#;Divisor register, most significant bits.
(synch:define UART: dlm[7:0] #x00 #f #f)
(synch:set! dll[7:0] (if (and (= uaddr[2:0] 0) uif-write dlab) udata[7:0] dll[7:0]))
(synch:set! dlm[7:0] (if (and (= uaddr[2:0] 1) uif-write dlab) udata[7:0] dlm[7:0]))
(synch:define UART: divisor-reload #t #f #f
	      (and (= uaddr[2:0] 0) uif-write dlab))

#;Receive buffer register.
(synch:define UART: rbr[7:0] #x00 #f #f 0)
#;Interrupt Identification register.
(synch:define UART: iir[7:0] #x01 #f #f 0)
#;Shadow MODEM status register.
(synch:define UART: msr-shadow[7:0] #x10 #f #f
	      (if (and (= uaddr[2:0] 6) uif-read)
		  (BAUDx16 msr[7:0])
		  msr-shadow[7:0]))
#;Line status register.
(synch:define UART: lsr[7:0] #x60 #f #f 0)

(define-synchronous-system uif)

;;;================================================================
;;;			    BAUDx16 Generator
;;;================================================================

#;This block contains the BAUDx16 Generator.
(synch:register-block 'BAUDx16 'xin-clock)
(define baudx16-reset- #t)
#;Crystal input frequency.
(defvar xin-clock 3.6864e6)

#;Divisor
(synch:defmacro UART:baudx16-divisor (logior (ash (uif dlm[7:0]) 8)
					     (uif dll[7:0])))
#;Was divisor changed?
(synch:define UART: divisor-reloaded #t #f #f (uif divisor-reload))
#;Was divisor changed last cycle?
(synch:define UART: divisor-loaded #t #f #f divisor-reloaded)
#;Reload baudx16-counter.
(synch:define UART: baud-reload #t #f #f
	      (and divisor-reloaded (not divisor-loaded)))

#;Divide xin-clock by baudx16-divisor.
(synch:define UART: baudx16-counter[15:0] *xxxx* #f #f
	      (if (or baud-reload (zero? baudx16-counter[15:0]))
		  baudx16-divisor
		  (- baudx16-counter[15:0] 1)))

#;BAUDx16 rate x16 if divisor by > 1.
(synch:define UART: baudx16-clock #f #f #f
	      (not (or baud-reload (zero? baudx16-counter[15:0]))))
#;BAUDx16 rate x16.
(synch:defmacro UART:trx-clock
		(if (>= 1 baudx16-divisor)
		    xin-clock
		    baudx16-clock))
(synch:defcheck (if (>= 1 baudx16-divisor)
		    (trx-sim 1)))

#;MODEM status register.
(synch:define UART: msr[7:0] #x10 #f #f
	      (logior
	       (ash (logand #xf (logxor msr[7:0] (uif msr-shadow[7:0]))) 4)
	       (if (logbit? 4 (uif mcr[7:0])) ;Loopback!
		   (booleans-to-number uart:rts uart:dtr uart:out1 uart:out2)
		   (booleans-to-number uart:cts uart:dsr uart:ri uart:dcd))))

(define-synchronous-system BAUDx16
  (synch:print
   "baud-ld" baud-reload
   "counter" (byte1 baudx16-counter[15:0]) "counter" (byte0 baudx16-counter[15:0])
   "baud-clk" baudx16-clock
   "trx-clk" trx-clock
   ))

;;;================================================================
;;;		       Serial Transmit/Receive
;;;================================================================

#;This block contains the TRX Generator.
(synch:register-block 'trx)
(define trx-reset- #t)
#;BAUD frequency x 16.
(defvar trx-clock #t)

#;No parity bit.
(defconst parity-none  #b000)
#;Odd parity.
(defconst parity-odd   #b001)
#;Even parity.
(defconst parity-even  #b011)
#;Space parity.
(defconst parity-space #b101)
#;Mark parity.
(defconst parity-mark  #b111)

#;Local (registered) copy of LCR.
(synch:define UART: llcr[7:0] 0 #f #f (uif lcr[7:0]))
#;Parity field from LCR.
(synch:defmacro UART:parity-mode[2:0] (bit-field llcr[7:0] 4 7))

;;;================================================================
;;;			Transmit state-machine
;;;================================================================

#;BAUD rate counter for transmit
(synch:define UART: tx-clk-cnt[3:0] *xxxx* #f #f
	      (if (zero? tx-clk-cnt[3:0]) 15 (- tx-clk-cnt[3:0] 1)))
#;BAUD rate clock enable for transmit
(synch:define UART: tx-step #f #f #f (zero? tx-clk-cnt[3:0]))

#;Marking (quiescent).
(defconst tx-marking 0)
#;Spacing for start-bit.
(defconst tx-start 1)
#;Sending data.
(defconst tx-char 3)
#;Sending Parity.
(defconst tx-parity 2)

#;Serial transmit state-machine
(synch:define UART: tx-state[1:0] tx-marking #f #f
	      (if tx-step
		  (casev tx-state[1:0]
		    ((,tx-marking)
		     (if (and (zero? stop-cnt[1:0]) (not tx-empty))
			 tx-start
			 tx-marking))
		    ((,tx-start) tx-char)
		    ((,tx-char) (if (zero? tx-bit-cnt[3:0])
				    (if (zero? parity-mode[2:0])
					tx-marking
					tx-parity)
				    tx-char))
		    ((,tx-parity) tx-marking))
		  tx-state[1:0]))

#;Data bit counter.
(synch:define UART: tx-bit-cnt[3:0] *xxxx* #f #f
	      (if tx-step
		  (casev tx-state[1:0]
		    ((,tx-marking) 0)
		    ((,tx-start) (+ 4 (logand #x3 llcr[7:0])))
		    ((,tx-char) (- tx-bit-cnt[3:0] 1))
		    ((,tx-parity) 0))
		  tx-bit-cnt[3:0]))
#;Transmit shift register.
(synch:define UART: tx-data[7:0] *xxxx* #f #f
	      (if tx-step
		  (casev tx-state[1:0]
		    ((,tx-start) (uif tbr[7:0]))
		    ((,tx-char) (ash tx-data[7:0] -1))
		    (else 0))
		  tx-data[7:0]))
#;Compute parity for serial data.
(synch:define UART: tx-parity-bit #f #f #f
	      (if tx-step
		  (casev tx-state[1:0]
		    ((,tx-char) (xor tx-parity-bit (logbit? 0 tx-data[7:0])))
		    (else #f))
		  tx-parity-bit))

#;Stop-bit counter
(synch:define UART: stop-cnt[1:0] 0 #f #f
	      (if tx-step
		  (casev tx-state[1:0]
		    ((,tx-marking) (- stop-cnt[1:0] 1))
		    ((,tx-char) (if (logbit? 2 llcr[7:0]) 1 0))
		    ((,tx-parity) stop-cnt[1:0])
		    ((,tx-start) 0))
		  stop-cnt[1:0]))

#;Serial output signal.
(synch:define UART: sout #t #f #t
	      (if (logbit? 6 llcr[7:0])
		  #f			;Break!
		  (if tx-step
		      (casev tx-state[1:0]
			((,tx-marking) #t)
			((,tx-start) #f)
			((,tx-char) (logbit? 0 tx-data[7:0]))
			((,tx-parity)
			 (casev parity-mode[2:0]
			   ((,parity-odd) tx-parity-bit)
			   ((,parity-even) (not tx-parity-bit))
			   ((,parity-mark) #t)
			   ((,parity-space) #t)
			   (else #f))))
		      sout)))

;;;================================================================
;;;			Receive state-machine
;;;================================================================

#;Serial input at pin.
;;(synch:defreginput UART: sin #t)
(synch:define INT: sin #t #f #f #t)
(synch:defshare sin UART:sin)
#;Serial input signal (includes loopback).
(synch:define UART: msin #t #f #f
	      (if (logbit? 4 (uif mcr[7:0]))
		  sout			;Loopback!
		  sin))
;;(synch:define UART: lmsin #t #f #f msin) ;delayed

#;BAUD rate counter for receive
(synch:define UART: rx-clk-cnt[3:0] *xxxx* #f #f
	      (casev rx-state[1:0]
		((,rx-marking)
		 (if (zero? rx-clk-cnt[3:0])
		     (if msin 7 0)	; Sample at middle of bit (1/2 of 15)
		     (- rx-clk-cnt[3:0] 1)))
		(else
		 (if (zero? rx-clk-cnt[3:0])
		     15
		     (- rx-clk-cnt[3:0] 1)))))

#;BAUD rate clock enable for receive
(synch:define UART: rx-step #f #f #f (zero? rx-clk-cnt[3:0]))

#;Marking (quiescent).
(defconst rx-marking 0)
#;Spacing for start-bit.
(defconst rx-start 1)
#;Sending data.
(defconst rx-char 3)
#;Sending Parity.
(defconst rx-parity 2)

#;Serial receive state-machine
(synch:define UART: rx-state[1:0] rx-marking #f #f
	      (if rx-step
		  (casev rx-state[1:0]
		    ((,rx-marking)
		     (if (and (zero? rx-clk-cnt[3:0]) (not msin))
			 rx-start
			 rx-marking))
		    ((,rx-start
		      ,rx-char
		      ,rx-parity)
		     (if (zero? rx-bit-cnt[3:0])
			 (if (zero? parity-mode[2:0])
			     rx-marking
			     rx-parity)
			 rx-char)))
		  rx-state[1:0]))

#;Data bit counter.
(synch:define UART: rx-bit-cnt[3:0] *xxxx* #f #f
	      (if rx-step
		  (casev rx-state[1:0]
		    ((,rx-marking) 0)
		    ((,rx-start) (+ 4 (logand #x3 llcr[7:0])))
		    ((,rx-char) (- rx-bit-cnt[3:0] 1))
		    ((,rx-parity) 0))
		  rx-bit-cnt[3:0]))
#;Receive shift register.
(synch:define UART: rx-data[7:0] *xxxx* #f #f
	      (if rx-step
		  (casev rx-state[1:0]
		    ((,rx-start) 0)
		    ((,rx-char) (logior
				 (if nsin
				     (ash 1 (+ 3 (logand #x3 llcr[7:0])))
				     0)
				 (ash rx-data[7:0] -1)))
		    (else rx-data[7:0]))
		  rx-data[7:0]))
#;Compute parity for serial data.
(synch:define UART: rx-parity-bit #f #f #f
	      (if rx-step
		  (casev rx-state[1:0]
		    ((,rx-char)
		     (casev parity-mode[2:0]
		       ((,parity-odd) (xor rx-parity-bit msin))
		       ((,parity-even) (xor rx-parity-bit msin))
		       ((,parity-mark) msin)
		       ((,parity-space) msin)
		       (else #f)))
		    (else #f))
		  rx-parity-bit))

#;Is received parity correct?
(synch:define UART: parity-error #f #f #f
	      (if rx-step
		  (casev rx-state[1:0]
		    ((,rx-marking) parity-error)
		    ((,rx-start) parity-error)
		    ((,rx-char) #f)
		    ((,rx-parity)
		     (casev parity-mode[2:0]
		       ((,parity-odd) rx-parity-bit)
		       ((,parity-even) (not rx-parity-bit))
		       ((,parity-mark) (not rx-parity-bit))
		       ((,parity-space) rx-parity-bit)
		       (else #f))))
		  rx-bit-cnt[3:0]))

(define-synchronous-system trx
  (synch:print
   "rx-data" rx-data[7:0]
   "tx-data" tx-data[7:0]
   ))

(solidify-database *board*)
(require 'run)

(define (stim:read addr)
  (set! uart:address addr)
  (set! uart:read-en #t)
  (uif-sim 2)
  (let ((urd (uif udata-out[7:0])))
    (set! uart:read-en #f)
    (uif-sim 1)
    urd))

(define (stim:write addr data)
  (set! uart:address addr)
  (set! uart:write-data data)
  (set! uart:write-en #t)
  (uif-sim 1)
  (set! uart:write-en #f)
  (uif-sim 1))

(uif-sim 1)
(BAUDx16-sim 10)
(stim:write 0 55)
(BAUDx16-sim 10)
(stim:write 1 03)
(BAUDx16-sim 10)
(stim:read 0)
(stim:read 1)
(BAUDx16-sim 10)
